@@ -1,5 +1,5 @@
 /*
- * $Id: com-socket.c,v 1.7 2002/05/02 13:01:08 mt Exp $
+ * $Id: com-socket.c,v 1.7.2.1 2003/05/07 11:14:08 mt Exp $
  *
  * Common functions for TCP/IP sockets
  *
@@ -30,7 +30,7 @@
  */
 
 #ifndef lint
-static char rcsid[] = "$Id: com-socket.c,v 1.7 2002/05/02 13:01:08 mt Exp $";
+static char rcsid[] = "$Id: com-socket.c,v 1.7.2.1 2003/05/07 11:14:08 mt Exp $";
 #endif
 
 #include <config.h>
@@ -97,6 +97,9 @@ static char rcsid[] = "$Id: com-socket.c,v 1.7 2002/05/02 13:01:08 mt Exp $";
 #if defined(HAVE_NET_IF_H)
 #  include <net/if.h>
 #endif
+#if defined(HAVE_NET_PFVAR_H)
+#  include <net/pfvar.h>
+#endif
 
 #if defined(HAVE_NETINET_IP_H)
 #   include <netinet/ip.h>
@@ -414,6 +417,7 @@ HLS *socket_init(int sock)
 
 	hls->kill = 0;
 	hls->ernr = 0;
+	hls->retr = 0;
 	hls->flag = 0;
 	hls->more = 0;
 	hls->ctyp = "HLS-TYPE";
@@ -994,7 +998,6 @@ int socket_exec(int timeout, int *close_flag)
 
 static void socket_ll_read(HLS *hls)
 {
-	static int retries = 0;
 	int len, cnt, nsock;
 	BUF *buf, *tmp;
 	struct sockaddr_in saddr;
@@ -1045,7 +1048,9 @@ static void socket_ll_read(HLS *hls)
 	if( (cnt=ioctl(hls->sock, FIONREAD, &len)) < 0) {
 		hls->ernr = errno;
 		syslog_error("can't get num of bytes: %s %d=%s",
-					hls->ctyp, hls->sock, hls->peer);
+		             hls->ctyp, hls->sock, hls->peer);
+		close(hls->sock);
+		hls->sock = -1;
 		return;
 	}
 #if defined(COMPILE_DEBUG)
@@ -1065,7 +1070,7 @@ static void socket_ll_read(HLS *hls)
 		** caused transfer aborts on bigger files.
 		** wait and retry before we assume this is a EOF.
 		*/
-		if((0 == cnt) && (++retries < MAX_RETRIES)) {
+		if((0 == cnt) && (++hls->retr < MAX_RETRIES)) {
 			syslog_write(T_DBG,
 			      "zero bytes to read reported: %s %d=%s",
 			      hls->ctyp, hls->sock, hls->peer);
@@ -1089,7 +1094,7 @@ static void socket_ll_read(HLS *hls)
 	/*
 	** else reset retry counter on success
 	*/
-	retries = 0;
+	hls->retr = 0;
 
 	/*
 	** Limit the receive buffer sizes
@@ -1972,15 +1977,21 @@ int        socket_chkladdr(u_int32_t addr)
 
 int socket_orgdst(HLS *phls, u_int32_t *addr, u_int16_t *port)
 {
-	struct sockaddr_in name;
+	struct sockaddr_in   name;
 #if defined(HAVE_LINUX_NETFILTER_IPV4_H) && defined(SO_ORIGINAL_DST)
-	struct sockaddr_in dest;
+	struct sockaddr_in   dest;
 #endif
 #if defined(HAVE_NETINET_IP_NAT_H) && defined(SIOCGNATL)
-	natlookup_t        natlook, *nlptr = &natlook;
-	int                rc, nat_fd = -1;
+	natlookup_t          natlook, *nlptr = &natlook;
+	int                  rc, nat_fd = -1;
 #endif
-	socklen_t          len;
+#if defined(HAVE_NET_PFVAR_H)
+	int                  rc, nat_fd = -1;
+	struct protoent      *proto;
+	struct sockaddr_in   peer;
+	struct pfioc_natlook natlook;
+#endif
+	socklen_t            len;
 
 	/*
 	** sanity args checks
@@ -1994,6 +2005,9 @@ int socket_orgdst(HLS *phls, u_int32_t *addr, u_int16_t *port)
 		syslog_error("can't get sockname for socket %d", phls->sock);
 		return -1;
 	}
+	syslog_write(T_DBG, "socket name address is %s:%d",
+		     socket_addr2str(ntohl(name.sin_addr.s_addr)),
+		                     ntohs(name.sin_port));
 
 #if defined(HAVE_LINUX_NETFILTER_IPV4_H) && defined(SO_ORIGINAL_DST)
 	/*
@@ -2020,7 +2034,7 @@ int socket_orgdst(HLS *phls, u_int32_t *addr, u_int16_t *port)
 		break;
 		default:
 			syslog_error(
-			"can't get iptables transparent destination");
+				"can't get iptables transparent destination");
 			return -1;
 		break;
 		}
@@ -2034,18 +2048,86 @@ int socket_orgdst(HLS *phls, u_int32_t *addr, u_int16_t *port)
 			return -1;
 		}
 
-		syslog_write(T_DBG,
-			"iptables transparent destination: %s:%d",
+		syslog_write(T_DBG, "iptables transparent destination: %s:%d",
 			socket_addr2str(ntohl(dest.sin_addr.s_addr)),
 			ntohs(dest.sin_port));
 
 		*addr = dest.sin_addr.s_addr;
 		*port = dest.sin_port;
-
 		return 0;
 	}
 #endif
 
+#if defined(HAVE_NET_PFVAR_H)
+
+	/*
+	** OpenBSD PF table lookup
+	*/
+	len = sizeof(peer);
+	memset(&peer, 0, len);
+	if(getpeername(phls->sock, (struct sockaddr *)&peer, &len) < 0) {
+		syslog_error("can't get peername for socket %d", phls->sock);
+		return -1;
+	}
+	syslog_write(T_DBG, "socket peer address is %s:%d",
+		     socket_addr2str(ntohl(peer.sin_addr.s_addr)),
+		                     ntohs(peer.sin_port));
+
+	if(!(proto = getprotobyname("tcp"))) {
+		syslog_error("can't get protocol number for tcp");
+		return -1;
+	}
+
+	if((nat_fd = open("/dev/pf", O_RDWR, 0)) < 0) {
+		endprotoent();
+		syslog_error("can't open pf device '/dev/pf'");
+		return -1;
+	}
+
+	memset(&natlook, 0, sizeof(natlook));
+	natlook.af              = AF_INET;
+	natlook.proto           = proto->p_proto;
+	natlook.direction       = PF_OUT;
+	natlook.saddr.v4.s_addr = peer.sin_addr.s_addr;		/* peer */
+	natlook.sport           = peer.sin_port;
+	natlook.daddr.v4.s_addr = name.sin_addr.s_addr;		/* sock */
+	natlook.dport           = name.sin_port;
+
+	endprotoent();
+
+	rc = ioctl(nat_fd, DIOCNATLOOK, &natlook);
+	close(nat_fd);
+
+	if(rc < 0) {
+		if(errno != ENOENT) {
+			syslog_error("can't get pfnat transparent destination");
+		}
+		return -1;
+	}
+
+#if defined(COMPILE_DEBUG)
+	debug(2, "pfnat transparent proxy destination: %s:%d",
+			socket_addr2str(ntohl(natlook.rdaddr.v4.s_addr)),
+			ntohs(natlook.rdport));
+#endif
+
+	if((natlook.rdport == name.sin_port) &&
+	   (natlook.rdaddr.v4.s_addr == name.sin_addr.s_addr))
+	{
+		syslog_write(T_DBG, "pfnat proxy destination %s:%d is local",
+		       socket_addr2str(ntohl(natlook.rdaddr.v4.s_addr)),
+		       ntohs(natlook.rdport));
+		return -1;
+	}
+	syslog_write(T_DBG, "pfnat transparent destination: %s:%d",
+	             socket_addr2str(ntohl(natlook.rdaddr.v4.s_addr)),
+	             ntohs(natlook.rdport));
+
+	*addr = natlook.rdaddr.v4.s_addr;
+	*port = natlook.rdport;
+	return 0;
+#endif
+
 #if defined(HAVE_NETINET_IP_NAT_H) && defined(SIOCGNATL)
 	/*
 	** BSD ipnat table lookup
@@ -2069,8 +2151,14 @@ int socket_orgdst(HLS *phls, u_int32_t *addr, u_int16_t *port)
 	rc = 0;
 	if(63 == (SIOCGNATL & 0xff)) {
 		rc = ioctl(nat_fd, SIOCGNATL, &nlptr);
+#if defined(COMPILE_DEBUG)
+		debug(2, "SIOCGNATL using &natlookptr: rc=%d", rc);
+#endif
 	} else {
 		rc = ioctl(nat_fd, SIOCGNATL, &natlook);
+#if defined(COMPILE_DEBUG)
+		debug(2, "SIOCGNATL using &natlook: rc=%d", rc);
+#endif
 	}
 	close(nat_fd);
 
@@ -2089,17 +2177,14 @@ int socket_orgdst(HLS *phls, u_int32_t *addr, u_int16_t *port)
 		       socket_addr2str(ntohl(natlook.nl_realip.s_addr)),
 		       ntohs(natlook.nl_realport));
 		return -1;
-	} else {
-		syslog_write(T_DBG,
-			"ipnat transparent destination: %s:%d",
-			socket_addr2str(ntohl(natlook.nl_realip.s_addr)),
-			ntohs(natlook.nl_realport));
-
-		*addr = natlook.nl_realip.s_addr;
-		*port = natlook.nl_realport;
-
-		return 0;
 	}
+	syslog_write(T_DBG, "ipnat transparent destination: %s:%d",
+		socket_addr2str(ntohl(natlook.nl_realip.s_addr)),
+		ntohs(natlook.nl_realport));
+
+	*addr = natlook.nl_realip.s_addr;
+	*port = natlook.nl_realport;
+	return 0;
 
 #else /* !BSD-IPNAT */
 
@@ -2187,6 +2272,10 @@ int getfqdomainname(char *fqdomain, size_t n)
 
 /* ------------------------------------------------------------
  * $Log: com-socket.c,v $
+ * Revision 1.7.2.1  2003/05/07 11:14:08  mt
+ * added OpenBSD pf-nat transparent proxy support
+ * fixed to use hls->retr instead of a static retry counter
+ *
  * Revision 1.7  2002/05/02 13:01:08  mt
  * merged with v1.8.2.2
  *
