@@ -1,5 +1,5 @@
 /*
- * $Id: ftp-client.c,v 1.9 2002/05/02 13:15:36 mt Exp $
+ * $Id: ftp-client.c,v 1.9.2.1 2003/05/07 11:11:33 mt Exp $
  *
  * Functions for the FTP Proxy client handling
  *
@@ -30,7 +30,7 @@
  */
 
 #ifndef lint
-static char rcsid[] = "$Id: ftp-client.c,v 1.9 2002/05/02 13:15:36 mt Exp $";
+static char rcsid[] = "$Id: ftp-client.c,v 1.9.2.1 2003/05/07 11:11:33 mt Exp $";
 #endif
 
 #include <config.h>
@@ -41,6 +41,7 @@ static char rcsid[] = "$Id: ftp-client.c,v 1.9 2002/05/02 13:15:36 mt Exp $";
 #  include <stdlib.h>
 #  include <stdarg.h>
 #  include <errno.h>
+#  include <ctype.h>
 #endif
 
 #if defined(HAVE_UNISTD_H)
@@ -81,6 +82,7 @@ static void client_cli_ctrl_read(char *str);
 static void client_srv_ctrl_read(char *str);
 static void client_srv_passive  (char *arg);
 static void client_xfer_fireup  (void);
+static int  client_setup_file(CONTEXT *ctx, char *who);
 
 
 /* ------------------------------------------------------------ */
@@ -305,36 +307,8 @@ void client_run(void)
 				memset(ctx.xfer_rep, 0,
 					sizeof(ctx.xfer_rep));
 			} else {
-				/*
-				** if an error occured and srv_data xfer
-				** is still running, try to ABORt it
-				**     else
-				** If we have outstanding server reply,
-				** it will come sooner or later. Then just
-				** mark it have to be passed to the client
-				** at the time it arrives.
-				*/
-				if( 0 != ctx.cli_data->ernr &&
-				   NULL != ctx.srv_data &&
-				   -1 != ctx.srv_data->sock)
-				{
-#if defined(MSG_OOB)
-					char str[4] = { IAC, IP, IAC, 0 };
-
-					socket_flag(ctx.srv_ctrl, MSG_OOB);
-					socket_write(ctx.srv_ctrl, str, 3);
-					socket_flag(ctx.srv_ctrl, 0);
-					str[0] = DM;
-					socket_write(ctx.srv_ctrl, str, 1);
-#endif
-					socket_printf(ctx.srv_ctrl,"ABOR\r\n");
-					client_respond(426, NULL,
-					"Connection closed; transfer aborted");
-					ctx.expect = EXP_ABOR;
-				} else {
-					if(ctx.expect == EXP_XFER)
-						ctx.expect = EXP_PTHR;
-				}
+				if(ctx.expect == EXP_XFER)
+					ctx.expect = EXP_PTHR;
 			}
 
 			/*
@@ -702,7 +676,7 @@ static void client_srv_ctrl_read(char *str)
 		*/
 		if (ctx.expect == EXP_CONN)
 			return;
-		if (ctx.expect == EXP_USER && NULL != ctx.userauth)
+		if (ctx.expect == EXP_USER && UAUTH_NONE != ctx.auth_mode)
 			return;
 
 #if defined(COMPILE_DEBUG)
@@ -742,7 +716,7 @@ static void client_srv_ctrl_read(char *str)
 				              ctx.username);
 				ctx.expect = EXP_USER;
 			} else {
-				if(NULL != ctx.userauth) {
+				if(UAUTH_NONE != ctx.auth_mode) {
 					client_respond(530, NULL,
 					               "Login incorrect");
 				} else {
@@ -761,13 +735,13 @@ static void client_srv_ctrl_read(char *str)
 			**	331=need password,
 			**	332=need password+account
 			*/
-			if(NULL != ctx.userauth) {
+			if(UAUTH_NONE != ctx.auth_mode) {
 				/*
 				** logged in, NO password needed
 				*/
 				if(c1 == 2 && c2 == 3) {
 					client_respond(230, NULL,
-					"User logged in, proceed.");
+					               "User logged in, proceed.");
 					ctx.expect = EXP_IDLE;
 					break;
 				} else
@@ -1103,7 +1077,8 @@ void client_reinit(void)
 	/*
 	** Remove the current user and status
 	*/
-	ctx.magic_auth = NULL;
+	ctx.auth_mode  = UAUTH_NONE;
+	ctx.magic_auth = 0;
 	if (ctx.userauth != NULL) {
 		misc_free(FL, ctx.userauth);
 		ctx.userauth = NULL;
@@ -1330,7 +1305,7 @@ void client_srv_open(void)
 **
 **	Function......:	client_setup
 **
-**	Parameters....:	pwd	client / user auth password
+**	Parameters....:	pwd	client / user password
 **
 **	Return........:	0 on success
 **
@@ -1341,11 +1316,11 @@ void client_srv_open(void)
 
 int client_setup(char *pwd)
 {
-	char *type;
+	char      *type;
+	char      *who;
 
 	/*
-	** Setup defaults for the client's DTP process;
-	** user specific setup is done in ldap_setup_user
+	** Setup defaults for the client's DTP process
 	*/
 	ctx.cli_mode = MOD_ACT_FTP;
 	ctx.cli_addr = ctx.cli_ctrl->addr;
@@ -1354,38 +1329,251 @@ int client_setup(char *pwd)
 	ctx.srv_port = INPORT_ANY;
 
 	/*
-	** check if we have to authenticate user
+	** select the proper name for user specific setup...
 	*/
-	if(NULL == (type = config_str(NULL, "UserAuthType", NULL))) {
-		/*
-		** setup use only
-		*/
-		return ldap_setup_user(&ctx, NULL);
+	if(NULL != ctx.userauth) {
+		who = ctx.userauth;
+	} else {
+		who = ctx.username;
 	}
 
 	/*
-	** authenticate + setup
+	** dont allow empty or invalid names...
 	*/
-/*
-	if( !strcasecmp(type, "pam")) {
-		if(0 == pam_auth_user(&ctx, pwd)) {
-			return ldap_setup_user(&ctx, NULL);
+	if(NULL != who && who[0]) {
+		char *ptr;
+		for(ptr=who; ptr[0]; ptr++)
+		if( !(isalnum(ptr[0]) || '-'  == ptr[0]  || '_'  == ptr[0])) {
+				syslog_write(U_ERR, "invalid user naem");
+			return -1;
 		}
-	} else
-*/
-	if( !strcasecmp(type, "ldap")) {
-		if(config_str(NULL, "LDAPServer", NULL) == NULL)
-			misc_die(FL, "client_setup: ?LDAPServer?");
-		return ldap_setup_user(&ctx, pwd);
+	} else {
+		syslog_write(U_ERR, "empty user name");
+		return -1;
 	}
 
-	misc_die(FL, "client_setup: ?UserAuthType?");
-	return -1; /* never reached */
+	/*
+	** user specific setup from config file
+	** with fallback to default values
+	*/
+	if(0 != client_setup_file(&ctx, who)) {
+		return -1;
+	}
+
+	/*
+	** check if we have to authenticate the user
+	**
+	** authenticate user and setup user specific
+	** from ldap server if configured
+	*/
+	type = config_str(NULL, "UserAuthType", NULL);
+	if(NULL != type) {
+
+		if(0 == strcasecmp(type, "ldap")) {
+			/*
+			** ldap server is mandatory
+			*/
+			if(NULL == config_str(NULL, "LDAPServer", NULL)) {
+				misc_die(FL, "client_setup: ?LDAPServer?");
+			}
+
+			/*
+			** ldap auth + setup
+			*/
+			if(0 != ldap_setup_user(&ctx, who, pwd ? pwd : ""))
+				return -1;
+		} else {
+			misc_die(FL, "client_setup: unknown ?UserAuthType?");
+		}
+
+	} else {
+		/*
+		** try ldap setup only
+		*/
+		ldap_setup_user(&ctx, who, NULL);
+	}
+
+	/*
+	** Evaluate mandatory settings or refuse to run.
+	*/
+	errno = 0;
+	if(INADDR_ANY == ctx.srv_addr || INADDR_BROADCAST == ctx.srv_addr) {
+		syslog_error("can't eval DestAddr for %s", ctx.cli_ctrl->peer);
+		return -1;
+	}
+	if(INPORT_ANY == ctx.srv_port) {
+		syslog_error("can't eval DestPort for %s", ctx.cli_ctrl->peer);
+		return -1;
+	}
+
+	return 0; /* all right */
+}
+
+
+/* ------------------------------------------------------------ **
+**
+**	Function......:	client_setup_file
+**
+**	Parameters....:	ctx	Pointer to user context
+**			who	Pointer to user name
+**
+**	Return........:	0 on success
+**
+**	Purpose.......: setup user-profile from config file
+**
+** ------------------------------------------------------------ */
+
+static int client_setup_file(CONTEXT *ctx, char *who)
+{
+	char      *p;
+	u_int16_t  l, u;
+
+	/*
+	** little bit sanity check
+	*/
+	if( !(ctx && who && *who)) {
+		return -1;
+	}
+
+	/*
+	** Inform the auditor that we are using the config file
+	*/
+	syslog_write(U_INF, "reading data for '%s' from cfg-file", who);
+
+	/*
+	** Evaluate DestinationAddress, except we have magic_addr
+	*/
+	if (INADDR_ANY != ctx->magic_addr) {
+		ctx->srv_addr = ctx->magic_addr;
+	} else {
+		ctx->srv_addr = config_addr(who, "DestinationAddress",
+		                                 INADDR_ANY);
+#if defined(COMPILE_DEBUG)
+		debug(2, "file DestAddr for %s: '%s'",
+		      ctx->cli_ctrl->peer, socket_addr2str(ctx->srv_addr));
+#endif
+	}
+
+	/*
+	** Evaluate DestinationPort, except we have magic_port
+	*/
+	if (INPORT_ANY != ctx->magic_port) {
+		ctx->srv_port = ctx->magic_port;
+	} else {
+		ctx->srv_port = config_port(who, "DestinationPort",
+		                                 IPPORT_FTP);
+#if defined(COMPILE_DEBUG)
+		debug(2, "file DestPort for %s: %d",
+		      ctx->cli_ctrl->peer, (int) ctx->srv_port);
+#endif
+	}
+
+	/*
+	** Evaluate the destination transfer mode
+	*/
+	p = config_str(who, "DestinationTransferMode", "client");
+	if(0 == strcasecmp(p, "active")) {
+		ctx->srv_mode = MOD_ACT_FTP;
+	} else
+	if(0 == strcasecmp(p, "passive")) {
+		ctx->srv_mode = MOD_PAS_FTP;
+	} else
+	if(0 == strcasecmp(p, "client")) {
+		ctx->srv_mode = MOD_CLI_FTP;
+	} else {
+		syslog_error("can't eval DestMode for %s",
+		             ctx->cli_ctrl->peer);
+		return -1;
+	}
+#if defined(COMPILE_DEBUG)
+	debug(2, "file DestMode for %s: %s", ctx->cli_ctrl->peer, p);
+#endif
+
+	/*
+	** Evaluate min/max destination port range
+	*/
+	l = config_port(who, "DestinationMinPort", INPORT_ANY);
+	u = config_port(who, "DestinationMaxPort", INPORT_ANY);
+	if (l > 0 && u > 0 && u >= l) {
+		ctx->srv_lrng = l;
+		ctx->srv_urng = u;
+	} else {
+		ctx->srv_lrng = INPORT_ANY;
+		ctx->srv_urng = INPORT_ANY;
+	}
+#if defined(COMPILE_DEBUG)
+	debug(2, "file DestRange for %s: %u-%u", ctx->cli_ctrl->peer,
+	         ctx->srv_lrng, ctx->srv_urng);
+#endif
+
+	/*
+	** Evaluate min/max active port range
+	*/
+	l = config_port(who, "ActiveMinDataPort", INPORT_ANY);
+	u = config_port(who, "ActiveMaxDataPort", INPORT_ANY);
+	if (l > 0 && u > 0 && u >= l) {
+		ctx->act_lrng = l;
+		ctx->act_urng = u;
+	} else {
+		/* do not try to bind a port < 1024 if running as UID != 0 */
+		if(0 == getuid()) {
+			ctx->act_lrng = (IPPORT_FTP - 1);
+			ctx->act_urng = (IPPORT_FTP - 1);
+		} else {
+			ctx->act_lrng = INPORT_ANY;
+			ctx->act_urng = INPORT_ANY;
+		}
+	}
+#if defined(COMPILE_DEBUG)
+	debug(2, "file ActiveRange for %s: %u-%u", ctx->cli_ctrl->peer,
+	         ctx->act_lrng, ctx->act_urng);
+#endif
+
+	/*
+	** Evaluate min/max passive port range
+	*/
+	l = config_port(who, "PassiveMinDataPort", INPORT_ANY);
+	u = config_port(who, "PassiveMaxDataPort", INPORT_ANY);
+	if (l > 0 && u > 0 && u >= l) {
+		ctx->pas_lrng = l;
+		ctx->pas_urng = u;
+	} else {
+		ctx->pas_lrng = INPORT_ANY;
+		ctx->pas_urng = INPORT_ANY;
+	}
+#if defined(COMPILE_DEBUG)
+	debug(2, "file PassiveRange for %s: %u-%u", ctx->cli_ctrl->peer,
+	         ctx->pas_lrng, ctx->pas_urng);
+#endif
+
+	/*
+	** Setup other configuration options
+	*/
+	ctx->same_adr = config_bool(who, "SameAddress", 1);
+	ctx->timeout  = config_int (who, "TimeOut",   900);
+#if defined(COMPILE_DEBUG)
+	debug(2, "file SameAddress for %s: %s", ctx->cli_ctrl->peer,
+	                                        ctx->same_adr ? "yes" : "no");
+	debug(2, "file TimeOut for %s: %d", ctx->cli_ctrl->peer, ctx->timeout);
+#endif
+
+	/*
+	** Adjust the allow/deny flags for the commands
+	*/
+	p = config_str(who, "ValidCommands", NULL);
+	cmds_set_allow(p);
+
+	return 0;
 }
 
 
 /* ------------------------------------------------------------
  * $Log: ftp-client.c,v $
+ * Revision 1.9.2.1  2003/05/07 11:11:33  mt
+ * removed broken ABOR sending on error while srv_data xfer still runs
+ * moved user config-profile reading from ftp-ldap to client_setup_file()
+ * changed to use new ctx.auth_mode flags to check if in user-auth mode
+ *
  * Revision 1.9  2002/05/02 13:15:36  mt
  * implemented simple (ldap based) user auth
  *
