@@ -1,5 +1,5 @@
 /*
- * $Id: ftp-cmds.c,v 1.10 2002/05/02 13:16:35 mt Exp $
+ * $Id: ftp-cmds.c,v 1.10.2.1 2003/05/07 11:10:45 mt Exp $
  *
  * FTP Proxy command handling
  *
@@ -30,7 +30,7 @@
  */
 
 #ifndef lint
-static char rcsid[] = "$Id: ftp-cmds.c,v 1.10 2002/05/02 13:16:35 mt Exp $";
+static char rcsid[] = "$Id: ftp-cmds.c,v 1.10.2.1 2003/05/07 11:10:45 mt Exp $";
 #endif
 
 #include <config.h>
@@ -101,6 +101,13 @@ static void cmds_epsv(CONTEXT *ctx, char *arg);
 #endif
 
 
+/* ------------------------------------------------------------ */
+
+static int  parse_magic_dest(CONTEXT *ctx, char *dest);
+static int  parse_magic_user(CONTEXT *ctx, char *uarg,
+                             char a_sep, int a_first,
+                             char u_sep, int u_force);
+
 /* ------------------------------------------------------------ */
 
 #if defined(HAVE_REGEX)
@@ -345,9 +352,7 @@ static void cmds_pthr(CONTEXT *ctx, char *arg)
 
 static void cmds_user(CONTEXT *ctx, char *arg)
 {
-	CMD                *cmd;
-	char               *p, *q;
-	int                need_auth;
+	CMD  *cmd;
 
 	if (ctx == NULL)		/* Basic sanity check	*/
 		misc_die(FL, "cmds_user: ?ctx?");
@@ -373,6 +378,7 @@ static void cmds_user(CONTEXT *ctx, char *arg)
 	*/
 	cmds_set_allow(config_str(NULL, "ValidCommands", NULL));
 	for (cmd = cmdlist; cmd->name != NULL; cmd++) {
+		char *p;
 		if (strcasecmp("USER", cmd->name) != 0)
 			continue;
 		if (cmd->regex == NULL)
@@ -442,122 +448,120 @@ static void cmds_user(CONTEXT *ctx, char *arg)
 	}
 
 	/*
-	** Check if we should use the auth-user mode
+	** Check if we need auth and should use the auth-user mode
 	*/
-	need_auth = 0;
 	if(NULL != config_str(NULL, "UserAuthType", NULL)) {
-		need_auth = 1;
+		ctx->auth_mode = UAUTH_FTP;
 		ctx->magic_auth = config_str(NULL, "UserAuthMagic", NULL);
+		if(NULL != ctx->magic_auth) {
+			if( sizeof("auth") != strlen(ctx->magic_auth)) {
+				syslog_write(T_ERR, "invalid UserAuthMagic");
+				client_respond(530, NULL, "Not logged in");
+				client_reinit();
+				return;
+			}
+			if(strncasecmp(ctx->magic_auth, "auth", sizeof("auth")-1))
+				ctx->auth_mode = UAUTH_MUA; /* user%auth */
+			else
+				ctx->auth_mode = UAUTH_MAU; /* auth%user */
+		}
 	}
 
 	/*
 	** Check if we have to parse 'auth' user name...
 	*/
-	if(need_auth && (NULL != ctx->magic_auth)) {
+	if(NULL != ctx->magic_auth) {
+		int   is_ok = 1;
+		char  a_sep = ctx->auth_mode == UAUTH_MAU
+		            ? ctx->magic_auth[sizeof("auth")-1]
+		            : ctx->magic_auth[0];
 #if defined(COMPILE_DEBUG)
-		debug(2, "parsing '%s' using auth-magic '%s'",
+		debug(2, "parsing '%s' using auth-magic '%.512s'",
 			arg, ctx->magic_auth);
 #endif
+
+		if(config_bool(NULL, "ForceMagicUser", 0) != 0) {
+			char *u_sep = config_str(NULL, "UserMagicChar",
+			              config_str(NULL, "UseMagicChar", "@"));
+			is_ok = parse_magic_user(ctx, arg,   a_sep,
+			                         ctx->auth_mode == UAUTH_MAU,
+			                         u_sep[0], 1);
+		} else
+		if(config_bool(NULL, "AllowMagicUser", 0) != 0) {
+			char *u_sep = config_str(NULL, "UserMagicChar",
+			              config_str(NULL, "UseMagicChar", "@"));
+			is_ok = parse_magic_user(ctx, arg,   a_sep,
+			                         ctx->auth_mode == UAUTH_MAU,
+			                         u_sep[0], 0);
+		} else {
+			is_ok = parse_magic_user(ctx, arg,   a_sep,
+			                         ctx->auth_mode == UAUTH_MAU,
+			                         0,        0);
+		}
+
+		if(is_ok || NULL == ctx->userauth || NULL == ctx->username ||
+		            '\0' == ctx->userauth || '\0' == ctx->username) {
+			if(1 == is_ok) {
+				syslog_write(U_ERR,
+				             "missed magic dest in 'USER' from %s",
+				             ctx->cli_ctrl->peer);
+			} else {
+				syslog_write(U_ERR,
+				             "invalid magic in 'USER' from %s",
+				             ctx->cli_ctrl->peer);
+			}
+			client_respond(530, NULL, "Not logged in");
+			client_reinit();
+			return;
+		}
+	} else {
 		/*
-		** Check how to parse for a client auth-name...
-		** type1: auth@user[@host:port] or
-		** type2: user@auth[@host:port] (NcFTP type5)
+		** USER="user[<u_sep>host[:port]]"
 		*/
-		if(sizeof("auth") == strlen(ctx->magic_auth) &&
-                   !strncasecmp(ctx->magic_auth, "auth", sizeof("auth")-1))
-		{
-			if((q = strchr(arg, ctx->magic_auth[sizeof("auth")-1])))
-			{
-				*q++ = '\0';
-				if( strlen(arg)) {
-					ctx->userauth = misc_strdup(FL, arg);
+		if(config_bool(NULL, "ForceMagicUser", 0) != 0) {
+			char *p, *u_sep = config_str(NULL, "UserMagicChar",
+			                  config_str(NULL, "UseMagicChar", "@"));
+			if( (p = strrchr(arg, u_sep[0]))) {
+				*p++ = '\0';
+				if(-1 == parse_magic_dest(ctx, p)) {
+					syslog_write(U_ERR,
+					             "invalid magic in 'USER' from %s",
+					             ctx->cli_ctrl->peer);
+					client_respond(530, NULL,
+					               "Not logged in");
+					client_reinit();
+					return;
 				}
-				arg  = q;
+			} else {
+				syslog_write(U_ERR,
+					"magic dest missed in 'USER' from %s",
+					ctx->cli_ctrl->peer);
+				client_respond(530, NULL, "Not logged in");
+				client_reinit();
+				return;
 			}
 		} else
-		if((q = strchr(arg, ctx->magic_auth[0]))) {
-			/*
-			** @auth may be followed by @host:port
-			*/
-			if(config_bool(NULL, "AllowMagicUser", 0) &&
-			   (p = config_str(NULL, "UseMagicChar", "@")) &&
-			   (p = strchr(q+1, p[0])))
-			{
-				if(p>q+1) {
-					ctx->userauth = misc_alloc(FL, p-q);
-					misc_strncpy(ctx->userauth, q+1, p-q);
-				}
-				while(*p) *q++=*p++;
-				*q = '\0';
-			} else {
-				*q++ = '\0';
-				if( strlen(q)) {
-					ctx->userauth = misc_strdup(FL, q);
+		if(config_bool(NULL, "AllowMagicUser", 0) != 0) {
+			char *p, *u_sep = config_str(NULL, "UserMagicChar",
+			                  config_str(NULL, "UseMagicChar", "@"));
+			if( (p = strrchr(arg, u_sep[0]))) {
+				*p++ = '\0';
+				if(-1 == parse_magic_dest(ctx, p)) {
+					syslog_write(U_ERR,
+					             "invalid magic in 'USER' from %s",
+					             ctx->cli_ctrl->peer);
+					client_respond(530, NULL,
+					               "Not logged in");
+					client_reinit();
+					return;
 				}
 			}
 		}
-
-		syslog_write(T_DBG, "parsed (magic)user='%s', auth='%s'",
-		                    NIL(arg), NIL(ctx->userauth));
-
-		/*
-		** Check if non-empty auth is given
-		*/
-		if(NULL == ctx->userauth) {
-			client_respond(501, NULL, "Permission denied");
-			syslog_write(U_WRN, "'USER' without auth from %s",
-			                    ctx->cli_ctrl->peer);
-			return;
-		}
-
-		/*
-		** Check for the user name - again...
-		*/
-		if (arg == NULL || *arg == '\0') {
+		ctx->username = misc_strdup(FL, arg);
+		if(NULL == ctx->username || '\0' == ctx->username) {
 			client_respond(501, NULL, "Missing user name");
 			syslog_write(U_WRN, "'USER' without name from %s",
 			                    ctx->cli_ctrl->peer);
-			client_reinit();
-			return;
-		}
-	}
-
-
-	/*
-	** Check for permission and existence of "magic"
-	** destination address and port information
-	*/
-	if(config_bool(NULL, "AllowMagicUser", 0) != 0 &&
-           (p = config_str(NULL, "UseMagicChar", "@")) &&
-	   (p = strchr(arg, *p)) != NULL)
-	{
-		*p++ = '\0';
-		if ((q = strchr(p, ':')) != NULL) {
-			*q++ = '\0';
-			ctx->magic_port = socket_str2port(q, IPPORT_FTP);
-		} else {
-			ctx->magic_port = IPPORT_FTP;
-		}
-		ctx->magic_addr = socket_str2addr(p, INADDR_ANY);
-		if (ctx->magic_addr == INADDR_ANY ||
-		    ctx->magic_port == INPORT_ANY) {
-			client_respond(501, NULL,
-				"Invalid destination in user name");
-			syslog_write(U_WRN,
-				"bad 'USER@' dest '%s:%s' from %s",
-				NIL(p), NIL(q), ctx->cli_ctrl->peer);
-			client_reinit();
-			return;
-		}
-
-		/*
-		** Check for the user name - again...
-		*/
-		if (arg == NULL || *arg == '\0') {
-			client_respond(501, NULL, "Missing user name");
-			syslog_write(U_WRN, "'USER' without name from %s",
-			                    ctx->cli_ctrl->peer);
-			client_reinit();
 			return;
 		}
 	}
@@ -565,7 +569,9 @@ static void cmds_user(CONTEXT *ctx, char *arg)
 	/*
 	** Retrieve the relevant user information
 	*/
-	if (ctx->magic_addr != INADDR_ANY) {
+	if (ctx->magic_addr != INADDR_ANY &&
+	    ctx->magic_addr != INADDR_NONE)
+	{
 		syslog_write(U_INF, "'USER %s' dest %s:%d from %s",
 				arg, socket_addr2str(ctx->magic_addr),
 				(int)ctx->magic_port, ctx->cli_ctrl->peer);
@@ -580,20 +586,7 @@ static void cmds_user(CONTEXT *ctx, char *arg)
 				arg, ctx->cli_ctrl->peer);
 	}
 
-	/*
-	** remember ftp user name
-	*/
-	ctx->username = misc_strdup(FL, arg);
-
-	/*
-	** copy normal ftp user name into userauth
-	** if we need any auth but have no userauth
-	*/
-	if(need_auth && (NULL == ctx->userauth)) {
-		ctx->userauth = misc_strdup(FL, arg);
-	}
-
-	if(NULL != ctx->userauth) {
+	if(UAUTH_NONE != ctx->auth_mode) {
 		/*
 		** anable PASS command only...
 		*/
@@ -604,8 +597,7 @@ static void cmds_user(CONTEXT *ctx, char *arg)
 		** PASS+auth as well, because it may be needed
 		** for auth itself and for profile reading...
 		*/
-		client_respond(331, NULL,
-		               "User name okay, need password.");
+		client_respond(331, NULL, "User name okay, need password.");
 	} else {
 		/*
 		** read user's profile, connect the server
@@ -613,6 +605,10 @@ static void cmds_user(CONTEXT *ctx, char *arg)
 		if(0 == client_setup(NULL)) {
 			client_srv_open();
 		} else {
+			/*
+			** FIXME: client_respond required? checkit!!
+			*/
+			client_respond(530, NULL, "Not logged in");
 			client_reinit();
 		}
 	}
@@ -634,7 +630,7 @@ static void cmds_user(CONTEXT *ctx, char *arg)
 
 static void cmds_pass(CONTEXT *ctx, char *arg)
 {
-	char *pass, *q;
+	char *pass = NULL, *q = NULL;
 
 	if (ctx == NULL)		/* Basic sanity check	*/
 		misc_die(FL, "cmds_pass: ?ctx?");
@@ -645,52 +641,57 @@ static void cmds_pass(CONTEXT *ctx, char *arg)
 	syslog_write(U_INF, "'PASS XXXX' from %s", ctx->cli_ctrl->peer);
 
 	/*
-	** Check if have to parse auth-pass... may be NULL
+	** should never be NULL, but ...
+	** if no password supplied, send none either
 	*/
-	pass = 0;
-	if(NULL != ctx->magic_auth) {
-		/*
-		** Check how to parse for a client auth pass...
-		** type1: auth@pass or type2: pass@auth (NcFTP type5)
-		*/
-		if(NULL != arg && sizeof("auth") == strlen(ctx->magic_auth) &&
-		   !strncasecmp(ctx->magic_auth, "auth", sizeof("auth")-1))
-		{
-			if((q = strchr(arg, ctx->magic_auth[sizeof("auth")-1])))
-			{
-				*q++ = '\0';
-				if(strlen(arg)) {
-					pass = arg;
-				}
-				if( strlen(q)) {
-					ctx->userpass = misc_strdup(FL, q);
-				}
-			}
-		} else {
-			if((NULL != arg) &&
-			   (q = strrchr(arg, ctx->magic_auth[0])))
-			{
-				*q++ = '\0';
-				if( strlen(q)) {
-					pass = q;
-				}
-				if(strlen(arg)) {
-					ctx->userpass = misc_strdup(FL, arg);
-				}
-			}
-		}
-	} else {
-		/* we may use normal ftp pwd */
+	if(NULL == arg)
+		pass = "";
+	else
 		pass = arg;
-	}
 
 	/*
-	** check if we are in auth mode
+	** Check if we are in auth mode...
 	*/
-	if(NULL != ctx->userauth) {
+	if(UAUTH_NONE != ctx->auth_mode) {
 		/*
-		** OK, we have all data to auth user, read
-		** his proxy-profile and connect to server
+		** Check if we should do auth using
+		** the normal FTP PASS comand.
+		*/
+		if(UAUTH_FTP == ctx->auth_mode) {
+			ctx->userpass = misc_strdup(FL, pass);
+		} else
+		/*
+		** Check if have to parse for magic
+		** auth pass in the FTP PASS command.
+		*/
+		if(NULL != ctx->magic_auth && pass[0] != '\0') {
+			if(ctx->auth_mode == UAUTH_MAU) {
+				q = strchr(pass, ctx->magic_auth[sizeof("auth")-1]);
+				if(NULL != q) {
+					*q++ = '\0';
+					ctx->userpass = misc_strdup(FL, q);
+				}
+			} else {
+				q = strrchr(pass, ctx->magic_auth[0]);
+				if(NULL != q) {
+					*q++ = '\0';
+					ctx->userpass = misc_strdup(FL, pass);
+					pass          = q;
+				}
+			}
+			if(NULL == q) {
+				syslog_write(U_ERR,
+				             "invalid magic in 'PASS' from %s",
+				             ctx->cli_ctrl->peer);
+				client_respond(530, NULL, "Not logged in");
+				client_reinit();
+				return;
+			}
+		}
+
+		/*
+		** OK, we have all data to auth user, read his
+		** proxy-profile (if any) and connect to server
 		*/
 		if(0 == client_setup(pass)) {
 			client_srv_open();
@@ -698,10 +699,7 @@ static void cmds_pass(CONTEXT *ctx, char *arg)
 			client_respond(530, NULL, "Not logged in");
 			client_reinit();
 		}
-		return;
-
 	} else {
-
 		/*
 		** paranoia check...
 		*/
@@ -712,16 +710,10 @@ static void cmds_pass(CONTEXT *ctx, char *arg)
 			return;
 		}
 
-		/*
-		** If no password is supplied, send none either
-		*/
-		if (arg == NULL)
-			arg = "";
-
 		/*
 		** Send to server, but do not display
 		*/
-		socket_printf(ctx->srv_ctrl, "PASS %.1024s\r\n", arg);
+		socket_printf(ctx->srv_ctrl, "PASS %.1024s\r\n", pass);
 		syslog_write(U_INF, "'PASS XXXX' from %s",
 		             ctx->cli_ctrl->peer);
 
@@ -1402,11 +1394,224 @@ char *cmds_reg_exec(void *regex, char *str)
 	*/
 	return NULL;
 }
+#endif /* !HAVE_REGEX */
+
+static int parse_magic_user(CONTEXT *ctx, char *uarg,
+                            char a_sep, int a_first,
+                            char u_sep, int u_force)
+{
+	char *p, *q;
+
+	if(NULL == uarg || '\0' == uarg || '\0' == a_sep) {
+		misc_die(FL, "parse_magic_user: ?uarg? ?a_sep?");
+	}
+
+#if defined(COMPILE_DEBUG)
+	debug(2, "parse_magic_user: uarg='%.512s' as=%c af=%d us=%c uf=%d",
+	         uarg, a_sep, a_first, 0 == u_sep? '0' : u_sep, u_force);
 #endif
 
+	if('\0' == u_sep) {
+		if(a_first) {
+			/*
+			** USER="auth<a_sep>user"
+			*/
+			p = strchr(uarg, a_sep);
+			if(NULL == p) {
+#if defined(COMPILE_DEBUG)
+				debug(3, "parse_magic_user: a_sep => NULL");
+#endif
+				return -1;
+			}
+			*p++ = '\0';
+			if('\0' == p[0] || '\0' == uarg) {
+#if defined(COMPILE_DEBUG)
+				debug(3, "parse_magic_user: a_sep => ''");
+#endif
+				return -1;
+			}
+			ctx->userauth = misc_strdup(FL, uarg);
+			ctx->username = misc_strdup(FL, p);
+		} else {
+			/*
+			** USER="user<a_sep>auth"
+			*/
+			p = strrchr(uarg, a_sep);
+			if(NULL == p) {
+#if defined(COMPILE_DEBUG)
+				debug(3, "parse_magic_user: a_sep => NULL");
+#endif
+				return -1;
+			}
+			*p++ = '\0';
+			if('\0' == p[0] || '\0' == uarg) {
+#if defined(COMPILE_DEBUG)
+				debug(3, "parse_magic_user: a_sep => ''");
+#endif
+				return -1;
+			}
+			ctx->username = misc_strdup(FL, uarg);
+			ctx->userauth = misc_strdup(FL, p);
+		}
+#if defined(COMPILE_DEBUG)
+		debug(2, "magic user='%.256s' auth='%.256s'",
+		         NIL(ctx->username), NIL(ctx->userauth));
+#endif
+		return 0;
+	}
+
+	if(a_first) {
+		/*
+		** USER="auth<a_sep>user[<u_sep>host[:port]]"
+		*/
+		p = strchr(uarg, a_sep);
+		if(NULL == p) {
+#if defined(COMPILE_DEBUG)
+			debug(3, "parse_magic_user: a_sep => NULL");
+#endif
+			return -1;
+		}
+		*p++ = '\0';
+		if('\0' == p[0] || '\0' == uarg) {
+#if defined(COMPILE_DEBUG)
+			debug(3, "parse_magic_user: a_sep => ''");
+#endif
+			return -1;
+		}
+		q = strrchr(p, u_sep);
+		if(NULL == q) {
+#if defined(COMPILE_DEBUG)
+			debug(3, "parse_magic_user: u_sep => NULL, user => '%.512s'", p);
+#endif
+			if(u_force)
+				return  1;
+		} else {
+			*q++ = '\0';
+			if('\0' == p[0] || '\0' == q[0]) {
+#if defined(COMPILE_DEBUG)
+				debug(3, "parse_magic_user: u_sep => '', user => '%.512s'", p);
+#endif
+				return -1;
+			}
+			if(-1 == parse_magic_dest(ctx, q))
+				return -1;
+		}
+		ctx->userauth = misc_strdup(FL, uarg);
+		ctx->username = misc_strdup(FL, p);
+#if defined(COMPILE_DEBUG)
+		debug(2, "magic user='%.256s' auth='%.256s'",
+		         NIL(ctx->username), NIL(ctx->userauth));
+#endif
+		return 0;
+	}
+
+	/*
+	** USER="user<a_sep>auth[<u_sep>host[:port]]"
+	*/
+	p = strrchr(uarg, a_sep);
+	if(NULL == p) {
+#if defined(COMPILE_DEBUG)
+		debug(3, "parse_magic_user: a_sep => NULL");
+#endif
+		return -1;
+	}
+	*p++ = '\0';
+	if('\0' == p[0] || '\0' == uarg) {
+#if defined(COMPILE_DEBUG)
+		debug(3, "parse_magic_user: a_sep => ''");
+#endif
+		return -1;
+	}
+	if(a_sep == u_sep) {
+		q = strrchr(uarg, u_sep);
+		if(NULL == q) {
+#if defined(COMPILE_DEBUG)
+			debug(3, "parse_magic_user: u_sep => NULL, user => '%.512s', dest => '%.512s'", uarg, p);
+#endif
+			if(u_force)
+				return 1;
+			ctx->username = misc_strdup(FL, uarg);
+			ctx->userauth = misc_strdup(FL, p);
+		} else {
+			*q++ = '\0';
+			if('\0' == uarg[0] || '\0' == q[0]) {
+#if defined(COMPILE_DEBUG)
+				debug(3, "parse_magic_user: u_sep => '', user => '%.512s', auth => '%.512s', dest => '%.512s'", uarg, q, p);
+#endif
+				return -1;
+			}
+			if(-1 == parse_magic_dest(ctx, p))
+				return -1;
+			ctx->username = misc_strdup(FL, uarg);
+			ctx->userauth = misc_strdup(FL, q);
+		}
+	} else {
+		q = strchr(p, u_sep);
+		if(NULL == q) {
+#if defined(COMPILE_DEBUG)
+			debug(3, "parse_magic_user: u_sep => NULL, user => '%.512s', auth => '%.512s'", uarg, p);
+#endif
+			if(u_force)
+				return 1;
+		} else {
+			*q++ = '\0';
+			if('\0' == p[0] || '\0' == q[0]) {
+#if defined(COMPILE_DEBUG)
+				debug(3, "parse_magic_user: u_sep => '', user => '%.512s', auth => '%.512s', dest => '%.512s'", uarg, p, q);
+#endif
+				return -1;
+			}
+			if(-1 == parse_magic_dest(ctx, q))
+				return -1;
+		}
+		ctx->username = misc_strdup(FL, uarg);
+		ctx->userauth = misc_strdup(FL, p);
+	}
+#if defined(COMPILE_DEBUG)
+	debug(2, "magic user='%.256s' auth='%.256s'",
+	         NIL(ctx->username), NIL(ctx->userauth));
+#endif
+	return 0;
+}
+
+static int parse_magic_dest(CONTEXT *ctx, char *dest)
+{
+	char *ptr;
+
+	if(dest && dest[0]) {
+		if( (ptr = strrchr(dest, ':'))) {
+			*ptr++ = '\0';
+			ctx->magic_port = socket_str2port(ptr, IPPORT_FTP);
+		} else {
+			ctx->magic_port = IPPORT_FTP;
+		}
+		ctx->magic_addr = socket_str2addr(dest, INADDR_ANY);
+#if defined(COMPILE_DEBUG)
+		debug(2, "parse magic host='%.256s' port='%d'",
+		         socket_addr2str(ctx->magic_addr),
+		                         ctx->magic_port);
+#endif
+		if(ctx->magic_addr != INADDR_ANY &&
+		   ctx->magic_addr != INADDR_NONE) {
+			return 0;
+		}
+#if defined(COMPILE_DEBUG)
+	} else {
+		debug(2, "parse magic dest => NONE");
+#endif
+	}
+	return -1;
+}
+
 
 /* ------------------------------------------------------------
  * $Log: ftp-cmds.c,v $
+ * Revision 1.10.2.1  2003/05/07 11:10:45  mt
+ * - fixed user magic parsing to allow emal-address in
+ *   username while @ is used as user magic separator
+ * - added ForceUserMagic config variable to enforce
+ *   host[:port] presence in FTP USER command
+ *
  * Revision 1.10  2002/05/02 13:16:35  mt
  * implemented simple (ldap based) user auth
  *
