@@ -1,5 +1,5 @@
 /*
- * $Id: ftp-ldap.c,v 1.7 2002/05/02 13:17:12 mt Exp $
+ * $Id: ftp-ldap.c,v 1.7.2.2 2003/05/11 20:22:17 mt Exp $
  *
  * FTP Proxy LDAP interface handling
  *
@@ -30,7 +30,7 @@
  */
 
 #ifndef lint
-static char rcsid[] = "$Id: ftp-ldap.c,v 1.7 2002/05/02 13:17:12 mt Exp $";
+static char rcsid[] = "$Id: ftp-ldap.c,v 1.7.2.2 2003/05/11 20:22:17 mt Exp $";
 #endif
 
 #include <config.h>
@@ -88,45 +88,45 @@ static char rcsid[] = "$Id: ftp-ldap.c,v 1.7 2002/05/02 13:17:12 mt Exp $";
 /* ------------------------------------------------------------ */
 
 #if defined(HAVE_LIBLDAP)
-#  if defined(HAVE_LDAP_UMICH)
-#    if defined __sun__
-      /*
-       * there is only a forward definition of the LDAP
-       * connection handle struct in ldap.h on Solaris7,
-       * so we have no access to ld_errno.
-       */
-#      define GET_LDERROR(ld,le)   le = LDAP_OTHER
-#    else
-#      if defined(LDAP_OPT_ERROR_NUMBER)
-         /*
-         ** OpenLDAP 2.x
-         */
-#        define GET_LDERROR(ld,le) \
-                ldap_get_option(ld,LDAP_OPT_ERROR_NUMBER,&le)
-#      else
-         /*
-         ** UmichLDAP or OpenLDAP 1.x
-         */
-#        define GET_LDERROR(ld,le) le = (ld)->ld_errno
-#      endif
-#    endif
-#  else
-      /*
-      ** Netscape
-      */
-#    define GET_LDERROR(ld,le)     le = ldap_get_lderrno((ld), NULL, NULL)
-#  endif
+#   if defined(HAVE_LDAP_GET_LDERRNO)
+	/*
+	** Netscape
+	*/
+#       define GET_LDERROR(ld,le) le = ldap_get_lderrno(ld, NULL, NULL)
+
+#   elif defined(HAVE_LDAP_GET_OPTION) && defined(LDAP_OPT_ERROR_NUMBER)
+	/*
+	** OpenLDAP 2.x
+	*/
+#       define GET_LDERROR(ld,le) ldap_get_option(ld,LDAP_OPT_ERROR_NUMBER,&le)
+
+#   elif defined __sun__
+	/*
+	** there is only a forward declaration of the LDAP
+	** connection handle struct in ldap.h on Solaris7,
+	** so we have no access to ld_errno.
+	*/
+#	define GET_LDERROR(ld,le)   le = LDAP_OTHER
+
+#   else
+	/*
+	** UmichLDAP or OpenLDAP 1.x
+	*/
+#       define GET_LDERROR(ld,le) le = (ld)->ld_errno
+#   endif
 #endif
 
 
 /* ------------------------------------------------------------ */
 
 #if defined(HAVE_LIBLDAP)
-static int   ldap_fetch(CONTEXT *ctx, char *srv, char *who, char *pwd);
-static int   ldap_auth(LDAP *ld, LDAPMessage *e, char *who, char *pwd);
+static int   ldap_fetch(LDAP *ld, CONTEXT *ctx, char *who, char *pwd);
 static char *ldap_attrib(LDAP *ld, LDAPMessage *e, char *attr, char *dflt);
 static int   ldap_exists(LDAP *ld, LDAPMessage *e, char *attr,
                                                    char *vstr, int cs);
+static int   ldap_auth(LDAP *ld, LDAPMessage *e, char *who, char *pwd);
+static char* prep_bind_auto(LDAP *ld, char *flt, char *base, char *peer);
+static char* prep_bind_fmt(char *str, char *who);
 #endif
 
 /* ------------------------------------------------------------ **
@@ -134,197 +134,239 @@ static int   ldap_exists(LDAP *ld, LDAPMessage *e, char *attr,
 **	Function......:	ldap_setup_user
 **
 **	Parameters....:	ctx		Pointer to user context
+**			who		Pointer to user auth name
 **			pwd		Pointer to user auth pwd
 **
-**	Return........:	(none)
+**	Return........:	0 on success
 **
 **	Purpose.......: Read the user specific parameters from
-**			an LDAP Server if one is known. Else
-**			read the values from the config file.
+**			LDAP Server if one is known.
 **
 ** ------------------------------------------------------------ */
 
-int  ldap_setup_user(CONTEXT *ctx, char *pwd)
+int  ldap_setup_user(CONTEXT *ctx, char *who, char *pwd)
 {
-	char      *who, *p;
-	u_int16_t l, u;
-
-	if (ctx == NULL)		/* Basic sanity check	*/
-		misc_die(FL, "ldap_setup_user: ?ctx?");
+	char      *ptr = 0;
+	int        ver = 0;
 
 	/*
-	** remember the proper name we should use...
+	** avoid unused... compiler warnings
 	*/
-	if(NULL != ctx->userauth) {
-		who = ctx->userauth;
-	} else {
-		who = ctx->username;
-	}
+	ver = ver;
+	ptr = ptr;
+	pwd = pwd;
 
 	/*
-	** dont allow empty or invalid names...
+	** Basic sanity check
 	*/
-	pwd=pwd;
-	if( !(NULL != who && *who && !strchr(who, '*'))) {
-		syslog_error("invalid user name or with asterisk");
-		return -1;
-	}
+	if( !(ctx && who && *who))
+		misc_die(FL, "ldap_setup_user: ?ctx? ?who?");
 
 #if defined(HAVE_LIBLDAP)
+	/*
+	** use configured ldap version or prefer v3 since
+	** OpenLDAP 2.x library defaults to v2, but the
+	** server does not accept v2 binds per default...
+	*/
+#if   defined(LDAP_VERSION3)
+	ver = LDAP_VERSION3;
+#elif defined(LDAP_VERSION2)
+	ver = LDAP_VERSION2;
+#else
+	ver = 0;
+#endif
+	ptr = config_str(NULL, "LDAPVersion", NULL);
+	if(NULL != ptr) {
+		ver = atoi(ptr);
+	}
+
 	/*
 	** If an LDAP server is configured, insist on using it
 	*/
-	if ((p = config_str(NULL, "LDAPServer", NULL)) != NULL) {
-		int rc = ldap_fetch(ctx, p, who, pwd);
+	if((ptr = config_str(NULL, "LDAPServer", NULL)) != NULL) {
+		char       temp[MAX_PATH_SIZE];
+		char      *host;
+		u_int16_t  port;
+		int        rc;
+		LDAP      *ld;
+
+		misc_strncpy(temp, ptr, sizeof(temp));
 		/*
-		** check if we have to read profile from config...
+		** Determine LDAP server and port
 		*/
-		syslog_write(T_DBG, "ldap_fetch returned %d", rc);
-		if(0 != rc) return rc;
-	}
-#endif
-
-	/*
-	** Inform the auditor that we are using the config file
-	*/
-	syslog_write(U_INF, "reading data for '%s' from cfg-file", who);
-
-	/*
-	** Evaluate the destination FTP server address.
-	** This is mandatory! Refuse to run if none given.
-	*/
-	if (ctx->magic_addr != INADDR_ANY)
-		ctx->srv_addr = ctx->magic_addr;
-	else if ((ctx->srv_addr = config_addr(who,
-			"DestinationAddress",
-			INADDR_BROADCAST)) == INADDR_BROADCAST) {
-		errno = 0;
-		syslog_error("can't eval DestAddr for %s",
-		             ctx->cli_ctrl->peer);
-		return -1;
-	}
-#if defined(COMPILE_DEBUG)
-	debug(2, "DestAddr for %s: '%s'", ctx->cli_ctrl->peer,
-	         socket_addr2str(ctx->srv_addr));
-#endif
-
-	/*
-	** Evaluate the destination FTP server port
-	*/
-	if (ctx->magic_port != (u_int16_t)INPORT_ANY)
-		ctx->srv_port = ctx->magic_port;
-	else if ((ctx->srv_port = config_port(who,
-		 "DestinationPort", IPPORT_FTP)) == 0)
-	{
-		errno = 0;
-		syslog_error("can't eval DestPort for %s",
-					ctx->cli_ctrl->peer);
-		return -1;
-	}
-#if defined(COMPILE_DEBUG)
-	debug(2, "DestPort for %s: %d",
-			ctx->cli_ctrl->peer, (int) ctx->srv_port);
-#endif
-
-	/*
-	** Evaluate the destination transfer mode
-	*/
-	p = config_str(who, "DestinationTransferMode", "client");
-	if (strcasecmp(p, "active") == 0)
-		ctx->srv_mode = MOD_ACT_FTP;
-	else if (strcasecmp(p, "passive") == 0)
-		ctx->srv_mode = MOD_PAS_FTP;
-	else if (strcasecmp(p, "client") == 0)
-		ctx->srv_mode = MOD_CLI_FTP;
-	else {
-		syslog_error("can't eval DestMode for %s",
-					ctx->cli_ctrl->peer);
-		return -1;
-	}
-#if defined(COMPILE_DEBUG)
-	debug(2, "DestMode for %s: %s", ctx->cli_ctrl->peer, p);
-#endif
-
-	/*
-	** Evaluate the port ranges
-	*/
-	l = config_port(who, "DestinationMinPort", INPORT_ANY);
-	u = config_port(who, "DestinationMaxPort", INPORT_ANY);
-	if (l > 0 && u > 0 && u >= l) {
-		ctx->srv_lrng = l;
-		ctx->srv_urng = u;
-	} else {
-		ctx->srv_lrng = INPORT_ANY;
-		ctx->srv_urng = INPORT_ANY;
-	}
-#if defined(COMPILE_DEBUG)
-	debug(2, "DestRange for %s: %d-%d", ctx->cli_ctrl->peer,
-			(int) ctx->srv_lrng, (int) ctx->srv_urng);
-#endif
-
-	l = config_port(who, "ActiveMinDataPort", INPORT_ANY);
-	u = config_port(who, "ActiveMaxDataPort", INPORT_ANY);
-	if (l > 0 && u > 0 && u >= l) {
-		ctx->act_lrng = l;
-		ctx->act_urng = u;
-	} else {
-		/* do not try to bind a port < 1024 if running as UID != 0 */
-		if(0 == getuid()) {
-			ctx->act_lrng = (IPPORT_FTP - 1);
-			ctx->act_urng = (IPPORT_FTP - 1);
+		host = temp;
+		if(NULL != (ptr = strchr(temp, ':'))) {
+			*ptr++ = '\0';
+			port = (int) socket_str2port(ptr, LDAP_PORT);
 		} else {
-			ctx->act_lrng = INPORT_ANY;
-			ctx->act_urng = INPORT_ANY;
+			port = (int) LDAP_PORT;
 		}
+
+#if defined(COMPILE_DEBUG)
+		debug(2, "LDAP server: %s:%d", host, port);
+#endif
+
+		/*
+		** Ready to contact the LDAP server
+		*/
+		if((ld = ldap_init(host, port)) == NULL) {
+			syslog_write(T_ERR,
+			             "can't reach LDAP server %s:%u for %s",
+			             host, port, ctx->cli_ctrl->peer);
+			return -1;
+		} else {
+			syslog_write(T_DBG,
+			             "LDAP server %s:%u: initialized for %s",
+			              host, port, ctx->cli_ctrl->peer);
+		}
+
+		if(ver > 0) {
+#if defined(HAVE_LDAP_SET_OPTION) && defined(LDAP_OPT_PROTOCOL_VERSION)
+			ldap_set_option(ld, LDAP_OPT_PROTOCOL_VERSION, &ver);
+#else
+			ld->ld_version = ver;
+#endif
+		}
+
+		rc = ldap_fetch(ld, ctx, who, pwd);
+		ldap_unbind(ld);
+		return rc;
 	}
-#if defined(COMPILE_DEBUG)
-	debug(2, "ActiveRange for %s: %d-%d", ctx->cli_ctrl->peer,
-			(int) ctx->act_lrng, (int) ctx->act_urng);
 #endif
-
-	l = config_port(who, "PassiveMinDataPort", INPORT_ANY);
-	u = config_port(who, "PassiveMaxDataPort", INPORT_ANY);
-	if (l > 0 && u > 0 && u >= l) {
-		ctx->pas_lrng = l;
-		ctx->pas_urng = u;
-	} else {
-		ctx->pas_lrng = INPORT_ANY;
-		ctx->pas_urng = INPORT_ANY;
-	}
-#if defined(COMPILE_DEBUG)
-	debug(2, "PassiveRange for %s: %d-%d", ctx->cli_ctrl->peer,
-			(int) ctx->pas_lrng, (int) ctx->pas_urng);
-#endif
-
-	/*
-	** Setup other configuration options
-	*/
-	ctx->same_adr = config_bool(who, "SameAddress", 1);
-	ctx->timeout  = config_int (who, "TimeOut",   900);
-#if defined(COMPILE_DEBUG)
-	debug(2, "SameAddress for %s: %s", ctx->cli_ctrl->peer,
-				ctx->same_adr ? "yes" : "no");
-	debug(2, "TimeOut for %s: %d", ctx->cli_ctrl->peer,
-						ctx->timeout);
-#endif
-
-	/*
-	** Adjust the allow/deny flags for the commands
-	*/
-	p = config_str(who, "ValidCommands", NULL);
-	cmds_set_allow(p);
-
 	return 0;
 }
 
 
 #if defined(HAVE_LIBLDAP)
+/* ------------------------------------------------------------ **
+**
+**	Function......:	prep_bind_auto
+**
+**	Parameters....:	ld		Pointer ldap struct
+**			flt		LDAP search filter
+**			base		BaseDN to search in
+**			peer		Peer name for syslog
+**
+**	Return........:	bind-dn or NULL
+**
+**	Purpose.......:	Searches using anonymous bind
+**			for a dn we want to bind later.
+**
+** ------------------------------------------------------------ */
+
+static char* prep_bind_auto(LDAP *ld, char *flt, char *base, char *peer)
+{
+	LDAPMessage *result, *e;
+	char        *attrs[] = {0, 0};
+	char        *bind_dn, *p;
+	int          err;
+
+	if(NULL == base || '\0' == base[0]) {
+		misc_die(FL,"prep_bind_auto: ?base?");
+	}
+
+	/*
+	** pre-bind anonymously
+	*/
+	err = ldap_simple_bind_s(ld, 0, 0);
+	if(LDAP_SUCCESS != err) {
+		syslog_write(T_ERR,
+		       "can't bind LDAP anonymously for %s: %.512s",
+		       peer, ldap_err2string(err));
+		return NULL;
+	}
+
+	result   = 0;
+	attrs[0] = config_str(NULL, "LDAPIdentifier", "CN");
+
+	err = ldap_search_s(ld,  base,  LDAP_SCOPE_SUBTREE,
+	                    flt, attrs, 1, &result);
+	if(LDAP_SUCCESS != err) {
+		syslog_write(T_ERR,
+		       "can't find valid bind-dn for %s: %.512s",
+		       peer, ldap_err2string(err));
+		return NULL;
+	}
+
+	e = ldap_first_entry(ld, result);
+	if(NULL == e) {
+		GET_LDERROR(ld, err);
+		syslog_write(T_ERR,
+		       "can't find valid bind-dn for %s", peer);
+		return NULL;
+	}
+
+	/*
+	** OK, we have a DN
+	*/
+	if(NULL != (p = ldap_get_dn(ld, e))) {
+		bind_dn = misc_strdup(FL, p);
+		ldap_memfree(p);
+	} else {
+		bind_dn = NULL;
+	}
+	ldap_msgfree(result);
+	syslog_write(T_DBG, "auto bind-dn='%.256s' for %s",
+		             NIL(bind_dn), peer);
+	return bind_dn;
+}
+
+
+/* ------------------------------------------------------------ **
+**
+**	Function......:	prep_bind_fmt
+**
+**	Parameters....:	str		Bind-DN fmt string
+**			who		Current user name
+**
+**	Return........:	bind-dn or NULL
+**
+**	Purpose.......:	Checks if str contains a fmt
+**			and constructs a new base-dn.
+**
+** ------------------------------------------------------------ */
+
+static char* prep_bind_fmt(char *str, char *who)
+{
+	char   *bind_dn, *p;
+	int    fmt = 0;
+	size_t len;
+
+	/*
+	** search for exactly one %s; if any other
+	** fmt's are present report an parse error.
+	*/
+	for(p=str; p && p[0] && (p = strchr(p, '%')); p++) {
+		if(!fmt && ('s' == p[1] || 'S' == p[1])) {
+			fmt = 1;
+		} else {
+			errno = 0;
+			misc_die(FL,"prep_bind_fmt: ?str?");
+		}
+	}
+
+	if(fmt) {
+		/*
+		** constuct LDAPBind DN and PW
+		*/
+		len = strlen(who) + strlen(str);
+		bind_dn = misc_alloc(FL, len);
+#if defined(HAVE_SNPRINTF)
+		snprintf(bind_dn, len, str, who);
+#else
+		sprintf(bind_dn, str, who);
+#endif
+		return bind_dn;
+	}
+	return 0;
+}
+
 /* ------------------------------------------------------------ **
 **
 **	Function......:	ldap_fetch
 **
-**	Parameters....:	ctx		Pointer to user context
-**			srv		Pointer ldap server name
+**	Parameters....:	ld		Pointer ldap struct
+**			ctx		Pointer to user context
 **			who		Pointer to user/auth name
 **			pwd		Pointer to user/auth pwd
 **
@@ -335,142 +377,29 @@ int  ldap_setup_user(CONTEXT *ctx, char *pwd)
 **
 ** ------------------------------------------------------------ */
 
-static int ldap_fetch(CONTEXT *ctx, char *srv, char *who, char *pwd)
+static int ldap_fetch(LDAP *ld, CONTEXT *ctx, char *who, char *pwd)
 {
-	char str[MAX_PATH_SIZE], *ptr, *p;
-	char *bind_dn, *bind_pw, *base, *idnt, *objc;
-	int port, lderr, is_fmt;
-	LDAP *ld;
-	LDAPMessage *result, *e;
+	char str[MAX_PATH_SIZE];
+	char *bind_dn, *bind_pw;
+	char *base_dn, *auth_dn;
+	char *idnt, *objc, *ptr, *p, *q;
+	char  lderr, auth_ok;
 	u_int16_t l, u;
+	LDAPMessage *result, *e;
 
 	/* Basic sanity */
-	if (ctx == NULL || srv == NULL || who == NULL)
-		misc_die(FL, "ldap_fetch: ?ctx? ?srv? ?who?");
-
-	/*
-	** prepare bind_dn: LDAPBindDN may contain
-	** _one_ %s (printf-fmt); replace it with
-	** the name passed via who parmeter...
-	*/
-	is_fmt  = 0;
-	bind_dn = NULL;
-	bind_pw = NULL;
-	if( (ptr = config_str(NULL, "LDAPBindDN", NULL))) {
-		size_t len;
-
-		/*
-		** search for exactly one %s; if any other
-		** fmt's are present report an parse error...
-		*/
-		for(p=ptr; p && p[0] && (p = strchr(p, '%')); p++) {
-			/* ignore %% pairs...? */
-			if('%' == p[1]) continue;
-			if(!is_fmt && ('s' == p[1] || 'S' == p[1])) {
-				is_fmt = 1;
-			} else {
-				errno = 0;
-				misc_die(FL, "ldap_fetch: ?BindDN-Fmt?");
-				return -1;
-			}
-		}
-
-		if( is_fmt) {
-			/*
-			** LDAPBindDN is a fmt-string
-			*/
-			len = strlen(who) + strlen(ptr);
-			bind_dn = misc_alloc(FL, len);
-#if defined(HAVE_SNPRINTF)
-			snprintf(bind_dn, len, ptr, who);
-#else
-			sprintf(bind_dn, ptr, who);
-#endif
-		} else {
-			bind_dn = misc_strdup(FL, ptr);
-		}
-	}
-
-	/*
-	** Determine LDAP server and port
-	*/
-	misc_strncpy(str, srv, sizeof(str));
-	if ((p = strchr(str, ':')) != NULL) {
-		*p++ = '\0';
-		port = (int) socket_str2port(p, LDAP_PORT);
-	} else {
-		port = (int) LDAP_PORT;
-	}
-
-#if defined(COMPILE_DEBUG)
-	debug(2, "LDAP server: %s:%d", str, port);
-#endif
-
-	/*
-	** Ready to contact the LDAP server (as anonymous user)
-	*/
-	if ((ld = ldap_init(str, port)) == NULL) {
-		errno = 0;
-		syslog_error("can't reach LDAP server %s:%d", str, port);
-		if(NULL != bind_dn)
-			misc_free(FL, bind_dn);
-		return -1;
-	} else {
-		syslog_write(T_DBG, "LDAP server %s:%d: init succeed",
-		                    str, port);
-	}
-
-	if((NULL != bind_dn)) {
-		if( !is_fmt) {
-			/*
-			** use LDAPBindPW for bind if not a fmt bind_dn
-			*/
-			bind_pw = config_str(NULL, "LDAPBindPW", NULL);
-		} else	bind_pw = pwd;
-	}
-
-	/*
-	** now, bind as anonymous or authenticated
-	*/
-	if(LDAP_SUCCESS != (lderr = ldap_simple_bind_s(ld, bind_dn, bind_pw)))
-	{
-		ldap_unbind(ld); /* needed? */
-		errno = 0;
-		if(NULL != bind_dn) {
-			syslog_error("can't bind LDAP using dn '%s'", bind_dn);
-			misc_free(FL, bind_dn);
-		} else {
-			syslog_error("can't bind LDAP anonymously");
-		}
-		return -1;
-	} else {
-		syslog_write(T_DBG,
-		             "LDAP bind dn='%.256s' pw='%.256s': succeed",
-		             NIL(bind_dn), NIL(bind_pw));
-	}
-	if(NULL != bind_dn) {
-		misc_free(FL, bind_dn);
-		bind_dn = 0;
+	if(ctx == NULL || ld == NULL || who == NULL) {
+		if(ld) ldap_unbind(ld);
+		misc_die(FL, "ldap_fetch: ?ctx? ?ld? ?who?");
 	}
 
 	/*
 	** construct filter for the search
 	**	(by LDAPIdentifier, maybe also ObjectClass)
 	*/
-	if ((objc = config_str(NULL, "LDAPObjectClass", "")) == NULL)
-		misc_die(FL, "ldap_fetch: ?ObjectClass?");
-	if ((idnt = config_str(NULL, "LDAPIdentifier", "")) == NULL)
-		misc_die(FL, "ldap_fetch: ?Identifier?");
-	if (*idnt == '\0')
-		idnt = "CN";
-	if (*objc == '\0') {
-#if defined(HAVE_SNPRINTF)
-		snprintf(str, sizeof(str), "(%.256s=%.256s)",
-			  idnt, who);
-#else
-		sprintf(str, "(%.256s=%.256s)", idnt, who);
-#endif
-	} else {
+	idnt = config_str(NULL, "LDAPIdentifier", "CN");
+	objc = config_str(NULL, "LDAPObjectClass", NULL);
+	if(NULL != objc) {
 #if defined(HAVE_SNPRINTF)
 		snprintf(str, sizeof(str),
 			  "(&(ObjectClass=%.256s)(%.256s=%.256s))",
@@ -479,88 +408,155 @@ static int ldap_fetch(CONTEXT *ctx, char *srv, char *who, char *pwd)
 		sprintf(str,
 			  "(&(ObjectClass=%.256s)(%.256s=%.256s))",
 			  objc,  idnt, who);
+#endif
+	} else {
+#if defined(HAVE_SNPRINTF)
+		snprintf(str, sizeof(str), "(%.256s=%.256s)",
+			  idnt, who);
+#else
+		sprintf(str, "(%.256s=%.256s)", idnt, who);
 #endif
 	}
 
+	auth_ok = 0; /* OK if non-anonymous bind */
+	auth_dn = config_str(NULL, "LDAPAuthDN", NULL);
+	base_dn = config_str(NULL, "LDAPBaseDN", NULL);
+	if(NULL != (ptr = config_str(NULL, "LDAPBindDN", NULL))) {
+		/*
+		** check if we should use auth-/base-dn for bind
+		*/
+		bind_pw = pwd;
+		bind_dn = 0;
+		if(0 == strcasecmp(ptr, "auto")) {
+			bind_dn = prep_bind_auto(ld, str, auth_dn ?
+			                         auth_dn : base_dn,
+			                         ctx->cli_ctrl->peer);
+			auth_ok = 1;
+			if(NULL == bind_dn) return -1;
+		} else
+		if(0 == strcasecmp(ptr, "AuthDN")) {
+			bind_dn = prep_bind_auto(ld, str, auth_dn,
+			                         ctx->cli_ctrl->peer);
+			auth_ok = 1;
+			if(NULL == bind_dn) return -1;
+		} else
+		if(0 == strcasecmp(ptr, "BaseDN")) {
+			bind_dn = prep_bind_auto(ld, str, base_dn,
+			                         ctx->cli_ctrl->peer);
+			auth_ok = 1;
+			if(NULL == bind_dn) return -1;
+		} else {
+			/*
+			** check if we have a format in BindDN
+			*/
+			bind_dn = prep_bind_fmt(ptr, who);
+			if(NULL == bind_dn) {
+				/*
+				** use static LDAPBind DN and PW
+				*/
+				bind_dn = misc_strdup(FL, ptr);
+				bind_pw = config_str(NULL, "LDAPBindPW", NULL);
+			} else {
+				auth_ok = 1;
+			}
+		}
 
-	syslog_write(U_INF, "reading data for '%s' from LDAP", who);
-	if ((base = config_str(NULL, "LDAPBaseDN", NULL)) != NULL)
-	{
-		syslog_write(T_DBG,
-		             "LDAP search: base='%.256s' filter='%.256s'",
-		             base, str);
-
-		if(LDAP_SUCCESS != (lderr = ldap_search_s(ld, base,
-	           LDAP_SCOPE_SUBTREE, str, NULL, 0, &result)))
-		{
-			errno = 0;
-			syslog_error("can't read LDAP data for %s: %.512s",
-			             ctx->cli_ctrl->peer,
-			             ldap_err2string(lderr));
-			ldap_unbind(ld);
+		/*
+		** bind usind a dn & pw
+		*/
+		lderr = ldap_simple_bind_s(ld, bind_dn, bind_pw);
+		if(LDAP_SUCCESS != lderr) {
+			syslog_write(U_ERR,
+			      "can't bind LDAP dn='%.256s' for %s: %.512s",
+			      bind_dn, ctx->cli_ctrl->peer,
+			      ldap_err2string(lderr));
 			return -1;
 		}
+		syslog_write(T_DBG,
+		             "LDAP bind to dn='%.256s': succeed", bind_dn);
+	} else {
+		/*
+		** bind anonymously
+		*/
+		lderr = ldap_simple_bind_s(ld, 0, 0);
+		if(LDAP_SUCCESS != lderr) {
+			syslog_write(T_ERR,
+			       "can't bind LDAP anonymously for %s: %.512s",
+			       ctx->cli_ctrl->peer, ldap_err2string(lderr));
+			return -1;
+		}
+	}
+
+	syslog_write(U_INF, "reading data for '%s' from LDAP", who);
+	if(NULL != base_dn) {
+		syslog_write(T_DBG,
+		             "LDAP search: base='%.256s' filter='%.256s'",
+		             base_dn, str);
+		result = 0;
+		lderr  = ldap_search_s(ld, base_dn, LDAP_SCOPE_SUBTREE,
+		                      str, NULL, 0, &result);
+		if(LDAP_SUCCESS != lderr) {
+			syslog_write(T_ERR,
+			             "can't read LDAP data for %s: %.512s",
+			             ctx->cli_ctrl->peer,
+			             ldap_err2string(lderr));
+			return -1;
+		}
+
+		
 
 		/*
 		** Check if we have a user data
 		** (else return 'error' or 'empty')
 		*/
-		if ((e = ldap_first_entry(ld, result)) == NULL)
-		{
+		if(NULL == (e = ldap_first_entry(ld, result))) {
 			GET_LDERROR(ld,lderr);
-			errno = 0;
-			syslog_error("can't read LDAP data for %s: %.512s",
-			             ctx->cli_ctrl->peer,
-			             ldap_err2string(lderr));
-			ldap_msgfree(result);
-			ldap_unbind(ld);
-			return -1;
+			syslog_write(T_DBG,
+			             "empty LDAP result for %s in base-dn='%s'",
+			             ctx->cli_ctrl->peer, base_dn);
+			e = result = NULL;
 		}
 	} else {
-		e = result = 0;
+		e = result = NULL;
 	}
 
 	/*
 	** Preform auth on userauth if type is ldap
 	*/
-	if( (ptr = config_str(NULL, "UserAuthType", NULL))
-	    && !strcasecmp(ptr, "ldap"))
-	{
+	ptr = config_str(NULL, "UserAuthType", NULL);
+	if( ptr && 0 == strcasecmp(ptr, "ldap")) {
 		LDAPMessage *res=0, *a=e;
+		int rc = 0;
+
 		/*
 		** if LDAPAuthDN set, do auth on a different base...
 		*/
-		if(NULL != (ptr = config_str(NULL, "LDAPAuthDN", NULL)))
-		{
+		if(NULL != auth_dn) {
 			syslog_write(T_DBG,
 			       "LDAP auth: base='%.256s' filter='%.256s'",
-			       ptr, str);
+			       auth_dn, str);
 
-			if(LDAP_SUCCESS != (lderr = ldap_search_s(ld,
-			   ptr, LDAP_SCOPE_SUBTREE, str, NULL, 0, &res)))
-			{
-				syslog_error("can't read LDAP auth-data"
-					     " for %s: %.512s",
-					     ctx->cli_ctrl->peer,
-					     ldap_err2string(lderr));
+			lderr = ldap_search_s(ld, auth_dn, LDAP_SCOPE_SUBTREE,
+			                      str, NULL, 0, &res);
+			if(LDAP_SUCCESS != lderr) {
+				syslog_write(T_ERR,
+				    "can't read LDAP auth-data for %s: %.512s",
+				    ctx->cli_ctrl->peer,
+				    ldap_err2string(lderr));
 				if(result) ldap_msgfree(result);
-				ldap_unbind(ld);
 				return -1;
 			}
 
 			if(NULL == (a = ldap_first_entry(ld, res))) {
 				GET_LDERROR(ld,lderr);
-				errno = 0;
-				syslog_error("can't read LDAP auth-data"
-					     " for %s: %.512s",
-					     ctx->cli_ctrl->peer,
-					     ldap_err2string(lderr));
+				syslog_write(T_WRN,
+				    "empty LDAP result for %s in auth-dn='%s'",
+				    ctx->cli_ctrl->peer, auth_dn);
 				if(result) ldap_msgfree(result);
-				ldap_msgfree(res);
-				ldap_unbind(ld);
 				return -1;
 			}
-		} else if(NULL == base || NULL == e) {
+		} else
+		if(NULL == base_dn || NULL == e) {
 			ldap_unbind(ld);
 			misc_die(FL, "ldap_fetch: ?LDAPBaseDN?");
 		}
@@ -568,188 +564,187 @@ static int ldap_fetch(CONTEXT *ctx, char *srv, char *who, char *pwd)
 		/*
 		** OK, let's check the user auth now
 		*/
-		if(0 != ldap_auth(ld, a, who, pwd)) {
-			errno = 0;
+		rc = ldap_auth(ld, a, who, pwd);
+		if(res) ldap_msgfree(res);
+		if(0 > rc)  {
 			syslog_write(U_ERR,
 			             "LDAP user auth failed for %s from %s",
 			             who, ctx->cli_ctrl->peer);
 			if(result) ldap_msgfree(result);
-			if(res)    ldap_msgfree(res);
-			ldap_unbind(ld);
 			return -1;
 		}
-		if(res) ldap_msgfree(res);
+		/*
+		** do not allow to configure UserAuthType=ldap
+		** and to skip all manual ldap checks without
+		** an sufficient ldap-bind.
+		*/
+		if(0 == rc && 0 == auth_ok) {
+			syslog_write(T_ERR, "LDAP auth config not sufficient");
+			if(result) ldap_msgfree(result);
+			return -1;
+		}
 	}
 
 	/*
-	** read proxy user profile data if we have a base
+	** read proxy user profile data ...
+	** if we have a base_dn and result
 	*/
-	if(NULL == base) {
-		ldap_unbind(ld);
+	if(NULL == base_dn || NULL == e) {
 		return 0;
 	}
 
 	/*
 	** Evaluate the destination FTP server address.
-	** This is mandatory! Refuse to run if none given.
 	*/
-	p = ldap_attrib(ld, e, "DestinationAddress", "255.255.255.255");
-	if (ctx->magic_addr != INADDR_ANY) {
-		ctx->srv_addr = ctx->magic_addr;
-	} else
-	if ((ctx->srv_addr = socket_str2addr(p,
-			INADDR_BROADCAST)) == INADDR_BROADCAST) {
-		errno = 0;
-		syslog_error("can't eval DestAddr for %s",
-					ctx->cli_ctrl->peer);
-		ldap_msgfree(result);
-		ldap_unbind(ld);
-		return -1;
-	}
+	p = ldap_attrib(ld, e, "DestinationAddress", NULL);
+	if(NULL != p && ctx->magic_addr == INADDR_ANY) {
+		ctx->srv_addr = socket_str2addr(p, INADDR_ANY);
+		if(INADDR_ANY == ctx->srv_addr) {
+			syslog_write(T_ERR, "can't eval DestAddr for %s",
+			                    ctx->cli_ctrl->peer);
+			ldap_msgfree(result);
+			return -1;
+		}
 #if defined(COMPILE_DEBUG)
-	debug(2, "DestAddr for %s: '%s'", ctx->cli_ctrl->peer,
+		debug(2, "ldap DestAddr for %s: '%s'", ctx->cli_ctrl->peer,
 				socket_addr2str(ctx->srv_addr));
 #endif
+	}
 
 	/*
 	** Evaluate the destination FTP server port
 	*/
-	p = ldap_attrib(ld, e, "DestinationPort", "ftp");
-	if (ctx->magic_port != (u_int16_t)INPORT_ANY) {
-		ctx->srv_port = ctx->magic_port;
-	} else
-	if ((ctx->srv_port = socket_str2port(p,
-			IPPORT_FTP)) == (u_int16_t)INPORT_ANY) {
-		errno = 0;
-		syslog_error("can't eval DestPort for %s",
-					ctx->cli_ctrl->peer);
-		ldap_msgfree(result);
-		ldap_unbind(ld);
-		return -1;
-	}
+	p = ldap_attrib(ld, e, "DestinationPort", NULL);
+	if(NULL != p && ctx->magic_port == INPORT_ANY) {
+		ctx->srv_port = socket_str2port(p, INPORT_ANY);
+		if(INPORT_ANY == ctx->srv_port) {
+			syslog_write(T_ERR, "can't eval DestPort for %s",
+			                    ctx->cli_ctrl->peer);
+			ldap_msgfree(result);
+			return -1;
+		}
 #if defined(COMPILE_DEBUG)
-	debug(2, "DestPort for %s: %d",
-			ctx->cli_ctrl->peer, (int) ctx->srv_port);
+		debug(2, "ldap DestPort for %s: %u",
+			ctx->cli_ctrl->peer, ctx->srv_port);
 #endif
+	}
 
 	/*
 	** Evaluate the destination transfer mode
 	*/
-	p = ldap_attrib(ld, e, "DestinationTransferMode", "client");
-	if (strcasecmp(p, "active") == 0)
-		ctx->srv_mode = MOD_ACT_FTP;
-	else if (strcasecmp(p, "passive") == 0)
-		ctx->srv_mode = MOD_PAS_FTP;
-	else if (strcasecmp(p, "client") == 0)
-		ctx->srv_mode = MOD_CLI_FTP;
-	else {
-		errno = 0;
-		syslog_error("can't eval DestMode for %s",
-					ctx->cli_ctrl->peer);
-		ldap_msgfree(result);
-		ldap_unbind(ld);
-		return -1;
-	}
+	p = ldap_attrib(ld, e, "DestinationTransferMode", NULL);
+	if(NULL != p) {
+		if(strcasecmp(p, "active") == 0)
+			ctx->srv_mode = MOD_ACT_FTP;
+		else if (strcasecmp(p, "passive") == 0)
+			ctx->srv_mode = MOD_PAS_FTP;
+		else if (strcasecmp(p, "client") == 0)
+			ctx->srv_mode = MOD_CLI_FTP;
+		else {
+			syslog_write(T_ERR, "can't eval DestMode for %s",
+			                    ctx->cli_ctrl->peer);
+			ldap_msgfree(result);
+			return -1;
+		}
 #if defined(COMPILE_DEBUG)
-	debug(2, "DestMode for %s: %s", ctx->cli_ctrl->peer, p);
+		debug(2, "ldap DestMode for %s: %s", ctx->cli_ctrl->peer, p);
 #endif
+	}
 
 	/*
 	** Evaluate the port ranges
 	*/
-	p = ldap_attrib(ld, e, "DestinationMinPort", "0");
-	l = socket_str2port(p, INPORT_ANY);
-	p = ldap_attrib(ld, e, "DestinationMaxPort", "0");
-	u = socket_str2port(p, INPORT_ANY);
-	if (l > 0 && u > 0 && u >= l) {
-		ctx->srv_lrng = l;
-		ctx->srv_urng = u;
-	} else {
-		ctx->srv_lrng = INPORT_ANY;
-		ctx->srv_urng = INPORT_ANY;
-	}
-#if defined(COMPILE_DEBUG)
-	debug(2, "DestRange for %s: %d-%d", ctx->cli_ctrl->peer,
-			(int) ctx->srv_lrng, (int) ctx->srv_urng);
-#endif
-
-	p = ldap_attrib(ld, e, "ActiveMinDataPort", "0");
-	l = socket_str2port(p, INPORT_ANY);
-	p = ldap_attrib(ld, e, "ActiveMaxDataPort", "0");
-	u = socket_str2port(p, INPORT_ANY);
-	if (l > 0 && u > 0 && u >= l) {
-		ctx->act_lrng = l;
-		ctx->act_urng = u;
-	} else {
-		/* do not try to bind a port < 1024 if running as UID != 0 */
-		if(0 == getuid()) {
-			ctx->act_lrng = (IPPORT_FTP - 1);
-			ctx->act_urng = (IPPORT_FTP - 1);
-		} else {
-			ctx->act_lrng = INPORT_ANY;
-			ctx->act_urng = INPORT_ANY;
+	p = ldap_attrib(ld, e, "DestinationMinPort", NULL);
+	q = ldap_attrib(ld, e, "DestinationMaxPort", NULL);
+	if(NULL != p && NULL != q) {
+		l = socket_str2port(p, INPORT_ANY);
+		u = socket_str2port(q, INPORT_ANY);
+		if (l > 0 && u > 0 && u >= l) {
+			ctx->srv_lrng = l;
+			ctx->srv_urng = u;
 		}
-	}
 #if defined(COMPILE_DEBUG)
-	debug(2, "ActiveRange for %s: %d-%d", ctx->cli_ctrl->peer,
-			(int) ctx->act_lrng, (int) ctx->act_urng);
+		debug(2, "ldap DestRange for %s: %u-%u", ctx->cli_ctrl->peer,
+		         ctx->srv_lrng, ctx->srv_urng);
 #endif
+	}
 
-	p = ldap_attrib(ld, e, "PassiveMinDataPort", "0");
-	l = socket_str2port(p, INPORT_ANY);
-	p = ldap_attrib(ld, e, "PassiveMaxDataPort", "0");
-	u = socket_str2port(p, INPORT_ANY);
-	if (l > 0 && u > 0 && u >= l) {
-		ctx->pas_lrng = l;
-		ctx->pas_urng = u;
-	} else {
-		ctx->pas_lrng = INPORT_ANY;
-		ctx->pas_urng = INPORT_ANY;
-	}
+	p = ldap_attrib(ld, e, "ActiveMinDataPort", NULL);
+	q = ldap_attrib(ld, e, "ActiveMaxDataPort", NULL);
+	if(NULL != p && NULL != q) {
+		l = socket_str2port(p, INPORT_ANY);
+		u = socket_str2port(q, INPORT_ANY);
+		if (l > 0 && u > 0 && u >= l) {
+			ctx->act_lrng = l;
+			ctx->act_urng = u;
+		}
 #if defined(COMPILE_DEBUG)
-	debug(2, "PassiveRange for %s: %d-%d", ctx->cli_ctrl->peer,
-			(int) ctx->pas_lrng, (int) ctx->pas_urng);
+		debug(2, "ActiveRange for %s: %u-%u", ctx->cli_ctrl->peer,
+		         ctx->act_lrng, ctx->act_urng);
 #endif
+	}
+
+	p = ldap_attrib(ld, e, "PassiveMinDataPort", NULL);
+	q = ldap_attrib(ld, e, "PassiveMaxDataPort", NULL);
+	if(NULL != p && NULL != q) {
+		l = socket_str2port(p, INPORT_ANY);
+		u = socket_str2port(q, INPORT_ANY);
+		if (l > 0 && u > 0 && u >= l) {
+			ctx->pas_lrng = l;
+			ctx->pas_urng = u;
+		}
+#if defined(COMPILE_DEBUG)
+		debug(2, "PassiveRange for %s: %u-%u", ctx->cli_ctrl->peer,
+		         ctx->pas_lrng, ctx->pas_urng);
+#endif
+	}
 
 	/*
 	** Setup other configuration options
 	*/
-	p = ldap_attrib(ld, e, "SameAddress", "yes");
-	if (strcasecmp(p, "y") == 0)
-		ctx->same_adr = 1;
-	else if (strcasecmp(p, "on") == 0)
-		ctx->same_adr = 1;
-	else if (strcasecmp(p, "yes") == 0)
-		ctx->same_adr = 1;
-	else if (strcasecmp(p, "true") == 0)
-		ctx->same_adr = 1;
-	else if (*p >= '0' && *p <= '9')
-		ctx->same_adr = (atoi(p) != 0);
-	else
-		ctx->same_adr = 0;
-	p = ldap_attrib(ld, e, "TimeOut", "900");
-	if (*p >= '0' && *p <= '9')
-		ctx->timeout = atoi(p);
-	else
-		ctx->timeout = 900;
+	p = ldap_attrib(ld, e, "SameAddress", NULL);
+	if(NULL != p) {
+		if (strcasecmp(p, "y") == 0)
+			ctx->same_adr = 1;
+		else if (strcasecmp(p, "on") == 0)
+			ctx->same_adr = 1;
+		else if (strcasecmp(p, "yes") == 0)
+			ctx->same_adr = 1;
+		else if (strcasecmp(p, "true") == 0)
+			ctx->same_adr = 1;
+		else if (*p >= '0' && *p <= '9')
+			ctx->same_adr = (atoi(p) != 0);
+		else
+			ctx->same_adr = 0;
 #if defined(COMPILE_DEBUG)
-	debug(2, "SameAddress for %s: %s", ctx->cli_ctrl->peer,
-				ctx->same_adr ? "yes" : "no");
-	debug(2, "TimeOut for %s: %d", ctx->cli_ctrl->peer,
-						ctx->timeout);
+		debug(2, "SameAddress for %s: %s", ctx->cli_ctrl->peer,
+		         ctx->same_adr ? "yes" : "no");
 #endif
+	}
+	p = ldap_attrib(ld, e, "TimeOut", "900");
+	if(NULL != p) {
+		if (*p >= '0' && *p <= '9')
+			ctx->timeout = atoi(p);
+		else
+			ctx->timeout = 900;
+#if defined(COMPILE_DEBUG)
+		debug(2, "TimeOut for %s: %d", ctx->cli_ctrl->peer,
+		         ctx->timeout);
+#endif
+	}
 
 	/*
 	** Adjust the allow/deny flags for the commands
 	*/
 	p = ldap_attrib(ld, e, "ValidCommands", NULL);
-	cmds_set_allow(p);
+	if(NULL != p) {
+		cmds_set_allow(p);
+	}
 
 	/*
 	** All relevant attributes have been evaluated
 	*/
 	ldap_msgfree(result);
-	ldap_unbind(ld);
 
 	return 0;
 }
@@ -775,6 +770,7 @@ static int   ldap_auth(LDAP *ld, LDAPMessage *e, char *who, char *pwd)
 	char str[MAX_PATH_SIZE];
 	char *v, *p, *q;
 	size_t len;
+	int    xrc = 0;
 
 	if (ld == NULL || e == NULL)
 		misc_die(FL, "ldap_checkauth: ?ld? ?e?");
@@ -797,6 +793,7 @@ static int   ldap_auth(LDAP *ld, LDAPMessage *e, char *who, char *pwd)
 				syslog_write(T_DBG,
 				"LDAP auth ok-check: '%.256s'='%.256s' passed",
 				NIL(str), NIL(v));
+				xrc = 1;
 			}
 		} else {
 			errno = 0;
@@ -839,7 +836,8 @@ static int   ldap_auth(LDAP *ld, LDAPMessage *e, char *who, char *pwd)
 			*/
 			len = sizeof("{crypt}")-1;
 			if(strncasecmp(q, "{crypt}", len)) {
-				syslog_error("ldap user auth - prefix missed");
+				syslog_write(T_ERR,
+				             "ldap user auth - prefix missed");
 				return -1;
 			}
 			q+=len;
@@ -876,7 +874,7 @@ static int   ldap_auth(LDAP *ld, LDAPMessage *e, char *who, char *pwd)
 			if(0 == strcmp(q, p)) {
 				syslog_write(T_DBG,
 				             "LDAP auth pw-check succeed");
-				return 0;
+				return xrc + 2;
 			}
 		}
 		syslog_write(T_DBG, "LDAP auth pw-check failed");
@@ -886,9 +884,9 @@ static int   ldap_auth(LDAP *ld, LDAPMessage *e, char *who, char *pwd)
 	}
 
 	/*
-	** OK, all configured LDAPAuth stuff succeed...
+	** OK, all configured manual LDAPAuth checks succeed...
 	*/
-	return 0;
+	return xrc;
 }
 
 
@@ -1017,6 +1015,14 @@ static int  ldap_exists(LDAP *ld, LDAPMessage *e, char *attr,
 
 /* ------------------------------------------------------------
  * $Log: ftp-ldap.c,v $
+ * Revision 1.7.2.2  2003/05/11 20:22:17  mt
+ * simplyfied ldap version handling
+ *
+ * Revision 1.7.2.1  2003/05/07 11:10:00  mt
+ * - moved user profile-config reading to ftp-client.c
+ * - added LDAP_VERSION handling with LDAPv3 default
+ * - improved user-auth to support auth via ldap-bind
+ *
  * Revision 1.7  2002/05/02 13:17:12  mt
  * implemented simple (ldap based) user auth
  *
